︠57d88ce4-b4c9-4784-92ce-0487b73fab98i︠
%html
<h1>Taylor Polynomials</h1>
<p>Since Taylor polynomials concentrate their accuracy around a point (making sure the function and polynomial have the same derivatives up to the $n$th derivative at that point), a Taylor polynomial may differ from a function significantly away from the point.</p>
<p>In the example below, we calculate and plot the Taylor polynomial $\hat{f}$ approximating the function $f(x)=1/x$ around the point $x=1$.&nbsp; We also print the error in the approximation at $x=3$, which is $|f(3)-\hat{f}(3)|$.&nbsp; Higher order Taylor polynomials usually give better approximations farther from the point, but in this case, as we increase the order of the approximation, the error gets much, much worse.&nbsp; Check this out by moving the slider to change the order of the approximation.</p>

︡9148b47c-4f99-45a5-8404-edc4e282ceaf︡{"html": "<h1>Taylor Polynomials</h1>\n<p>Since Taylor polynomials concentrate their accuracy around a point (making sure the function and polynomial have the same derivatives up to the $n$th derivative at that point), a Taylor polynomial may differ from a function significantly away from the point.</p>\n<p>In the example below, we calculate and plot the Taylor polynomial $\\hat{f}$ approximating the function $f(x)=1/x$ around the point $x=1$.&nbsp; We also print the error in the approximation at $x=3$, which is $|f(3)-\\hat{f}(3)|$.&nbsp; Higher order Taylor polynomials usually give better approximations farther from the point, but in this case, as we increase the order of the approximation, the error gets much, much worse.&nbsp; Check this out by moving the slider to change the order of the approximation.</p>"}︡
︠321ac6fe-2eac-4334-9fdf-6b85b41badd2︠
x0  = 1
f(x)= 1/x
p   = plot(f,-1,5, thickness=2)
dot = point((x0,f(x=x0)),pointsize=80,color='red')
@interact
def _(order=(1..12)):
  taylor = f.taylor(x,x0,order)
  taylor_plot = plot(taylor,-1, 5, color='green', thickness=2)
  error = abs(f(3)-taylor(3))
  html('$f(x)=%s$'%latex(f(x)))
  html('$\hat{f}(x;%s)=%s+\mathcal{O}(x^{%s})$'%(x0,latex(taylor(x)),order+1))
  html('|$f(3)-\hat{f}(3)|= %s$'%(error.n())) 
  show(dot + p + taylor_plot+line([(3,f(3)),(3,taylor(3))],color='purple',linestyle=':'), ymin=-50,ymax=50)
︡a85066a7-45b7-4d7d-8428-21f48374699d︡︡
︠e767fd27-1d75-4310-bc92-356b2faa312ei︠
%html
<h1>Lagrange Polynomials</h1>
<p>Instead of concentrating our accuracy around a specific point, let's make sure that our approximation function is exact at a number of specified points.&nbsp; The Lagrange polynomial for $n+1$ data points is the unique $n$th degree polynomial that passes through those points.&nbsp; We define the polynomial by taking linear combinations of functions so that we have the right values at the points we specify.</p>

︡5a10c557-337f-4c40-a5c7-fe6e31aad2f0︡{"html": "<h1>Lagrange Polynomials</h1>\n<p>Instead of concentrating our accuracy around a specific point, let's make sure that our approximation function is exact at a number of specified points.&nbsp; The Lagrange polynomial for $n+1$ data points is the unique $n$th degree polynomial that passes through those points.&nbsp; We define the polynomial by taking linear combinations of functions so that we have the right values at the points we specify.</p>"}︡
︠acda13b9-84f5-4032-853b-f5c6e84fedfb︠
pts=[(i,1/i) for i in [1..5]]
html.table(pts)
︡838246f2-27bf-496e-92a4-824dcd83a7a7︡{"html": "\n<div class=\"notruncate\">\n<table class=\"table_form\">\n<tbody>\n<tr class =\"row-a\">\n<td><span class=\"math\">1</span></td>\n<td><span class=\"math\">1</span></td>\n</tr>\n<tr class =\"row-b\">\n<td><span class=\"math\">2</span></td>\n<td><span class=\"math\">\\frac{1}{2}</span></td>\n</tr>\n<tr class =\"row-a\">\n<td><span class=\"math\">3</span></td>\n<td><span class=\"math\">\\frac{1}{3}</span></td>\n</tr>\n<tr class =\"row-b\">\n<td><span class=\"math\">4</span></td>\n<td><span class=\"math\">\\frac{1}{4}</span></td>\n</tr>\n<tr class =\"row-a\">\n<td><span class=\"math\">5</span></td>\n<td><span class=\"math\">\\frac{1}{5}</span></td>\n</tr>\n</tbody>\n</table>\n</div>\n"}︡
︠fb7b16e4-7d9c-44b8-9357-63316fbc3751︠
def switch_poly(pts,point_to_delete):
    xvalues=[p[0] for p in pts]
    deleted_point=xvalues.pop(point_to_delete)
    numerator=prod((x-pt) for pt in xvalues)
    denominator=numerator(x=deleted_point)
    return numerator/denominator
︡aa89f843-87b0-4783-a55f-6149ddbb3d0d︡︡
︠21ae7ab8-41b1-4bd3-976a-12814522a608i︠
%html
<p>For each $x$-coordinate, there is a polynomial that is 1 at that $x$-coordinate and zero at all of the others.&nbsp; Here is the polynomial that is zero at $x_0,x_1,x_3,x_4$ and 1 at $x_2$.</p>

︡cc030975-ff8f-4480-b55e-ac9b9c4812fb︡{"html": "<p>For each $x$-coordinate, there is a polynomial that is 1 at that $x$-coordinate and zero at all of the others.&nbsp; Here is the polynomial that is zero at $x_0,x_1,x_3,x_4$ and 1 at $x_2$.</p>"}︡
︠20165bef-adcc-4912-8d4d-c9237143b356︠
plot(switch_poly(pts,2),(x,0.8,5.2))
︡e23ff3ac-ef39-4743-931e-e532192e2469︡{"html": "<font color='black'><img src='cell://sage0.png'></font>"}︡
︠a535fcb8-4c83-409e-9b41-e54d2de71ba4i︠
%html
<p>Here is a plot of all of the polynomials, one for each $x_i$.</p>

︡e716f997-419d-4080-9052-aa33516f7987︡{"html": "<p>Here is a plot of all of the polynomials, one for each $x_i$.</p>"}︡
︠4fbbc217-bbdd-4282-892b-66b0a6c2320b︠
points([(p[0],1) for p in pts],color='red',pointsize=30)+plot([switch_poly(pts,i) for i in range(len(pts))],(x,0.8,5.2),fill=True)
︡1e642e3a-f4eb-425f-acb9-9f22e39760f5︡{"html": "<font color='black'><img src='cell://sage0.png'></font>"}︡
︠fc400e35-0089-471a-8f40-8298995faf3fi︠
%html
<p>The Lagrange polynomial is just a linear combination of each of switching polynomials.&nbsp; This guarantees that we have the correct values at the points $x_0$, $x_1$, $x_2$, $x_3$, and $x_4$.</p>

︡bc133fd8-e6a5-4fb1-bb10-4384a4f19f50︡{"html": "<p>The Lagrange polynomial is just a linear combination of each of switching polynomials.&nbsp; This guarantees that we have the correct values at the points $x_0$, $x_1$, $x_2$, $x_3$, and $x_4$.</p>"}︡
︠4c9413a1-9b65-48d0-a925-ae8a089b9cde︠
def lagrange_poly(pts):
    return sum(switch_poly(pts,i)*pts[i][1] for i in range(len(pts)))
︡29e9c442-0d9b-41f9-bd17-bf6943c9f423︡︡
︠d91f39a3-f605-4971-a219-4c1c9bb0820b︠
p(x)=lagrange_poly(pts)
p(x)
︡a7f67271-d01b-4136-bd40-dd5a22d3bf2f︡{"stderr": "Traceback (most recent call last):\n  File \"<stdin>\", line 1, in <module>\n  File \"_sage_input_3.py\", line 9, in <module>\n    exec compile(ur'open(\"___code___.py\",\"w\").write(\"# -*- coding: utf-8 -*-\\n\" + _support_.preparse_worksheet_cell(base64.b64decode(\"cCh4KT1sYWdyYW5nZV9wb2x5KHB0cykKcCh4KQ==\"),globals())+\"\\n\"); execfile(os.path.abspath(\"___code___.py\"))' + '\\n', '', 'single')\n  File \"\", line 1, in <module>\n    \n  File \"/tmp/tmpnv9_yB/___code___.py\", line 2, in <module>\n    __tmp__=var(\"x\"); p = symbolic_expression(lagrange_poly(pts)).function(x)\nNameError: name 'lagrange_poly' is not defined"}︡
︠9f2e429d-882a-4667-b933-7cc1b184a27b︠
p(x).expand()
︡34dbc57e-3678-4b09-a522-354b40845d65︡{"stderr": "Traceback (most recent call last):\n  File \"<stdin>\", line 1, in <module>\n  File \"_sage_input_9.py\", line 9, in <module>\n    exec compile(ur'open(\"___code___.py\",\"w\").write(\"# -*- coding: utf-8 -*-\\n\" + _support_.preparse_worksheet_cell(base64.b64decode(\"cCh4KS5leHBhbmQoKQ==\"),globals())+\"\\n\"); execfile(os.path.abspath(\"___code___.py\"))' + '\\n', '', 'single')\n  File \"\", line 1, in <module>\n    \n  File \"/tmp/tmpxKdKax/___code___.py\", line 2, in <module>\n    exec compile(ur'p(x).expand()' + '\\n', '', 'single')\n  File \"\", line 1, in <module>\n    \nNameError: name 'p' is not defined"}︡
︠87134705-14f1-4a60-9ea7-d45d230e7957i︠
%html
<p>Sage also has a builtin function to generate Lagrange polynomials as part of its polynomial objects.</p>

︡e0a7085d-15c6-4bac-99ed-c2e828ec55c0︡{"html": "<p>Sage also has a builtin function to generate Lagrange polynomials as part of its polynomial objects.</p>"}︡
︠af073082-e9b1-40bf-8264-35f0ebddd613︠
R = PolynomialRing(QQ, 'x')
L = R.lagrange_polynomial(pts)
L
︡6c67be41-9e30-4d26-a058-862237f10af7︡{"html": "<span class=\"math\">\\newcommand{\\Bold}[1]{\\mathbf{#1}}\\frac{1}{120} x^{4} - \\frac{1}{8} x^{3} + \\frac{17}{24} x^{2} - \\frac{15}{8} x + \\frac{137}{60}</span>"}︡
︠d8302b5e-2f4f-4b24-8806-79334a0347e2i︠
%html
<p>Note that Sage automatically evaluates a Polynomial object using Horner's form.&nbsp; Here we evaluate the polynomial at "$x$".&nbsp; A polynomial object also uses Horner form for evaluating a polynomial at a number.</p>

︡2359616c-8f7e-415a-8a9c-ad9e2d2f45f5︡{"html": "<p>Note that Sage automatically evaluates a Polynomial object using Horner's form.&nbsp; Here we evaluate the polynomial at \"$x$\".&nbsp; A polynomial object also uses Horner form for evaluating a polynomial at a number.</p>"}︡
︠ba6dc56e-83c8-46da-856e-514bd01fe4e4︠
L(x)
︡21cbf4bf-f8de-4f37-8a2c-0f9cc51dfef1︡{"html": "<span class=\"math\">\\newcommand{\\Bold}[1]{\\mathbf{#1}}\\frac{1}{120} \\, {\\left({\\left({\\left(x - 15\\right)} x + 85\\right)} x - 225\\right)} x + \\frac{137}{60}</span>"}︡
︠12a7cff0-5a6c-49e2-9eef-3615994fdd3di︠
%html
<p>Let's plot this approximation to see how good it is.</p>

︡cb00b4c5-27ff-401c-82c0-59ad38c83649︡{"html": "<p>Let's plot this approximation to see how good it is.</p>"}︡
︠ae5cdaa7-e651-40c2-9831-fc340288d0c9︠
x0  = 1
f(x)= 1/x
p   = plot(f,-1,5, thickness=2)
R = PolynomialRing(QQ, 'x')
@interact
def _(x_values=input_box([1,2,3,4,5])):
    pts=[(QQ(i),f(i)) for i in x_values]
    dots = point(pts,pointsize=80,color='red')
    L = R.lagrange_polynomial(pts)
    L_plot = plot(L,(x,-1, 5), color='green', thickness=2)
    error = abs(f(3)-L(3))
    html('$f(x)=%s$'%latex(f(x)))
    html('$L(x)=%s=%s$'%(latex(L(x)),latex(L(x).expand())))
    html('|$f(3)-L(3)|= %s$'%(latex(error.n()))) 
    show(dots + p + L_plot+line([(3,f(3)),(3,L(3))],color='purple',linestyle=':'), ymin=-10,ymax=10)
︡015f9757-0b8c-4699-a1c9-d05efe54ce2c︡︡
︠6ecd7480-5c4a-4cd0-bfab-0705d6315e72i︠
%html
<p>These approximations can be pretty bad for some functions.&nbsp; A famous example is the Runge function $$\frac{1}{1+25x^2}$$</p>

︡66d60371-f89a-4e21-a9ae-98fe8939b55a︡{"html": "<p>These approximations can be pretty bad for some functions.&nbsp; A famous example is the Runge function $$\\frac{1}{1+25x^2}$$</p>"}︡
︠bc50da19-ed51-4906-ab42-cb807e3dd324︠
plot(1/(1+25*x^2), (x,-2,2))
︡3ca381dc-8d49-4ad8-88f8-1e1c1e5db20d︡︡
︠0eb81504-7fcb-4a03-bf85-94ad49d83fbc︠
x0  = 1
f(x)= 1/(1+25*x^2)
x_bounds=(-2,2)
p   = plot(f,x_bounds, thickness=2)
R = PolynomialRing(QQ, 'x')
@interact
def _(x_values=input_box([-1,-7/8,..,1])):
    pts=[(QQ(i),f(i)) for i in x_values]
    dots = point(pts,pointsize=40,color='red')
    L = R.lagrange_polynomial(pts)
    L_plot = plot(L,x_bounds, color='green', thickness=2)
    error = abs(f(3)-L(3))
    html('$f(x)=%s$'%latex(f(x)))
    html('$L(x)=%s$'%(latex(L(x).expand())))
    html('|$f(3)-L(3)|= %s$'%(latex(error.n()))) 
    show(dots + p + L_plot, ymin=-10,ymax=10)
︡fb4bf823-e56c-4f18-8efc-0b884ef066ea︡︡
︠c73610e0-e6f4-4d97-8dee-fb83fdea9ed5i︠
%html
<p>Sage automatically evaluates a Polynomial object using a technique that is at least efficient as Horner's form (if there are lots of zero coefficients, then it is more efficient by constructing a graph; type "sage.rings.polynomial.polynomial_compiled.CompiledPolynomialFunction??" to see the details and code).&nbsp; Here we evaluate the polynomial at "$x$", and you can see that it looks like Horner's form is being used.</p>

︡921deaaf-d5f2-491f-81fb-0860c402b994︡{"html": "<p>Sage automatically evaluates a Polynomial object using a technique that is at least efficient as Horner's form (if there are lots of zero coefficients, then it is more efficient by constructing a graph; type \"sage.rings.polynomial.polynomial_compiled.CompiledPolynomialFunction??\" to see the details and code).&nbsp; Here we evaluate the polynomial at \"$x$\", and you can see that it looks like Horner's form is being used.</p>"}︡
